<?php 
/**
 * This importer copies plate inspections and their associated images from a Rigaku imager
 * into the LIMS for onward processing. It creates associated records in the LIMS, such as 
 * imagers and plate types, where needed.
 * 
 * Command-line arguments:
 * 
 * -h Show help
 * -d2016-02-28 - Begin import from this date (default date of last imported inspection)
 * -iRI54-0039 - Import only for this imager (default all imagers)
 * -b909s - (Re)import images for this barcode
 * -s5 - Skip this many inspections at the beginning (Default 0)
 * -m20 - Import at most this many inspections (default 10)
 * -l1 - Set logging level. 1=debug, 2=info, 3=warn, 4=error. (Default 2)
 * 
 * A note on time and time zones: It appears that RockMaker's database holds LOCAL, not UTC, times in 
 * DateImaged, with no offset information.  These times come (as I understand it) not from the database 
 * server but from the controller PC for the relevant imager. The controller PC may or may not be in the 
 * same time zone as the server (Oulu's RI2 controller was installed on New York time!) and may or may 
 * not observe DST; if it does observe DST, it may or may not switch at the same time as the server, 
 * which also may or may not observe it. Even if they all play by the same rules, their clocks may or may 
 * not be synchronised. All in all, this is an unholy mess. Realistically, the best you can do under the 
 * circumstances is to calculate the server's local GMT offset and apply that to DateImaged, hoping 
 * blindly that all the Formulatrix boxes have roughly the same idea of what time it is. For historical 
 * data this may be an hour (possibly two) adrift depending on how much fun DST happened to be that day. 
 * The normal operating scenario is to import imaging tasks shortly after they happen, so they will be 
 * mostly right (except right around the DST change, if there is one). In theory the DST change mean some
 * missed inspections once per year; under the circumstances, that's not the worst outcome. Re-importing 
 * the affected plate(s) with the -b option should overcome it.
 */

set_time_limit(0);
setUpClassAutoLoader();

define('IMAGINGSESSIONS_PER_RUN',10);
define('MINIMUM_INSPECTION_AGE_MINS',5);

//These are just defaults. Imagers are imported rarely, and these numbers can be changed later, 
//so no need for a separate config item.
define('IMAGERLOAD_ALERT_THRESHOLD_PERCENT',75);
define('IMAGERLOAD_WARNING_THRESHOLD_PERCENT',90);

define('IMAGER_NAME_UNKNOWN', 'Unknown Imager');

define('LOGLEVEL_DEBUG', 1);
define('LOGLEVEL_INFO', 2);
define('LOGLEVEL_WARN', 3);
define('LOGLEVEL_ERROR', 4);
$logLevel=LOGLEVEL_INFO; //may be overridden by argument

//Formulatrix: 0=visible, 1=UV. Probably additional values for SONICC, etc. Not defined in DB.
$lightPaths=array('Visible','UV');

$limitStart=0;
$limitTotal=IMAGINGSESSIONS_PER_RUN;
$fromDate=null;
$imager=null;
$barcode=null;
for($i=1;$i<count($argv);$i++){
	$arg=$argv[$i];
	if(preg_match('/^-h$/',$arg)){
		showHelp();
		exit();
	} else if(preg_match('/^-l[1-4]$/',$arg)){
		$logLevel=(int)substr($arg, 2);
	} else if(preg_match('/^-s\d+$/',$arg)){
		$limitStart=(int)substr($arg, 2);
	} else if(preg_match('/^-m\d+$/',$arg)){
		$limitTotal=(int)substr($arg, 2);
	} else if(preg_match('/^-i.+$/',$arg)){
		$imager=substr($arg, 2);
	} else if(preg_match('/^-b.+$/',$arg)){
		$barcode=substr($arg, 2);
	} else if(preg_match('/^-d\d\d\d\d-\d\d-\d\d+$/',$arg)){
		$fromDate=substr($arg,2);
	} else {
		
	}
}


//ini_set('mssql.secure_connection', 'on');

//Constants for Formulatrix' various imaging states. These are not defined in the RI or RM databases, you just have to know them.
// define('IMAGING_STATE_SCHEDULED',1);	//"Not completed" - planned but not yet due
// define('IMAGING_STATE_SKIPPED',2);		//"Skipped" - RockImager decided not to do it
// define('IMAGING_STATE_PENDING',3);		//"Pending" - Imaging task is overdue and will be done as soon as imager is free
// define('IMAGING_STATE_QUEUED',4);		//"Queued" - Imaging task is about to begin (plate is probably moving to camera)
// define('IMAGING_STATE_IMAGING',5);		//"Imaging" - Imaging task in progress, the plate is being imaged
// define('IMAGING_STATE_COMPLETED',6);	//"Completed" - Imaging task has been done. Last image has been taken, but plate may not be back in the hotel yet.
// define('IMAGING_STATE_CANCELLED',7);	//"Cancelled" - Imaging task was cancelled manually while in progress.


writeLog(LOGLEVEL_INFO,'Importer started');


writeLog(LOGLEVEL_DEBUG,'Attempting to get LIMS DB connection... ');
	database::connect();
writeLog(LOGLEVEL_DEBUG,'...got LIMS DB connection');

writeLog(LOGLEVEL_DEBUG,'Attempting to get LIMS session... ');
	session::init(new DummySession());
	session::set('isAdmin',true);
writeLog(LOGLEVEL_DEBUG,'...got LIMS session');


$sharedProjectId=null;


try {
	
	
	
	database::begin();	

	$user=user::getByName('rigakuimporter');
	if(empty($user)){
		$user=user::create(array(
				'name'=>'rigakuimporter',
				'fullname'=>'Rigaku Importer',
				'email'=>'rigimport@null.null',
				'password'=>'USELESSPASSWORD',
				'isactive'=>0
		));
		$user=$user['created'];
	}
	session::init(new DummySession());
	session::set('userId', $user['id']);
	session::set('isAdmin', true);
	
// 	$hostname=trim(config::get('rig_dbhost'));
// 	$port=trim(config::get('rig_dbport'));
// 	$username=trim(config::get('rig_dbuser'));
// 	$pw=trim(config::get('rig_dbpass'));
// 	$rmdbname=trim(config::get('rig_dbname'));
	
	foreach(PDO::getAvailableDrivers() as $driver){
		writeLog(LOGLEVEL_DEBUG, 'Database driver available: '.$driver);
	}
	
	$hostname='128.214.58.164';
	$port='1521';
	$username='ed-test';
	$pw='any-1010';
	$dbname='crystals';
	
	//Connect to Rigaku database
	writeLog(LOGLEVEL_DEBUG,'Attempting Rigaku DB connection... ');
	$rig=null;
	//$ri = new PDO ("oci:Server=$hostname, $port;Database=$dbname","$username","$pw");
	$ri = new PDO ("oci:Server=$hostname, $port","$username","$pw");
	$ri->setAttribute( PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION );
	writeLog(LOGLEVEL_DEBUG,'...got DB connection');
	
	//Find some common prerequisites
	
	$sharedProject=project::getByName('Shared');
	if(empty($sharedProject)){ throw new Exception('Could not find "Shared" project in LIMS'); }
	$sharedProjectId=$sharedProject['id'];
	
	$defaultProject=project::getByName('Default Project');
	if(empty($defaultProject)){ throw new Exception('Could not find "Default Project" in LIMS'); }
	$defaultProjectId=$defaultProject['id'];

	//$unknownUsernames is a comma-separated list of RockMaker usernames that are known to be defaults. 
	//They should not be plate owners, but often are because users forget to select themselves.
	//We assign their plates to username $defaultPlateOwner in the LIMS, and an admin has to sort it out.
	$unknownUsernames='_*** CHANGE THIS!!!! ***_, Default User, Administrator, FormulatrixAdmin';
	$defaultPlateOwnerUsername='unknownuser';
	$unknownUsernames=explode(',',$unknownUsernames);
	foreach($unknownUsernames as &$u){
		$u=trim($u);
	}
	$defaultPlateOwner=user::getByName($defaultPlateOwnerUsername);
	if(empty($defaultPlateOwner)){
		$defaultPlateOwner=user::create(array(
			'name'=>$defaultPlateOwnerUsername,
			'fullname'=>'Unknown User',
			'email'=>'unknownuser@null.null',
			'password'=>'USELESSPASSWORD',
			'isactive'=>0
		));
		$defaultPlateOwner=$defaultPlateOwner['created'];
	}
	if(empty($defaultPlateOwner)){
		throw new Exception('Could not find or create default user ['.$defaultPlateOwnerUsername.']');
	}
	$defaultPlateOwnerId=$defaultPlateOwner['id'];
	
	$scoringSystem=crystalscoringsystem::getByName('Custom');
	if(!$scoringSystem){
		$scoringSystem=crystalscoringsystem::getByName('Formulatrix');
		if(!$scoringSystem){
			$scoringSystem=importScoringSystem();
		}
	}
	if(!$scoringSystem){
		throw new Exception("Could not find or import Formulatrix scoring system");
	}
	
// 	$stmt=$rm->prepare("SELECT ID FROM CaptureProfileVersion");
// 	$stmt->execute(array());
// 	$ids=$stmt->fetchAll(PDO::FETCH_ASSOC);
// 	foreach($ids as $x){
// 		importCaptureProfileVersion($x['ID']);
// 	}
	
	database::commit();
	
	//Decide what to import

	if(null!=$barcode){
		
		//get plate
		$stmt=$rm->prepare('SELECT * FROM Plate WHERE Barcode=:barcode');
		$stmt->execute(array(':barcode'=>$barcode));
		$result=$stmt->fetch(PDO::FETCH_ASSOC);
		$fxPlateId=$result['ID'];
		
		//get its ExperimentPlate
		$stmt=$rm->prepare('SELECT * FROM ExperimentPlate WHERE PlateId=:plateid');
		$stmt->execute(array(':plateid'=>$fxPlateId));
		$result=$stmt->fetch(PDO::FETCH_ASSOC);
		$explateId=$result['ID'];
				
		//get the ImagingTasks
		//Time zone: Just interested in preventing very recent inspections from importing, so comparing (local) DateImaged to (local) GETDATE() is good enough.
		$stmt=$rm->prepare('SELECT * FROM ImagingTask WHERE ExperimentPlateId=:explateid AND DateImaged<=DATEADD(minute, -5, GETDATE()) ORDER BY DateImaged ASC');
		$stmt->execute(array(':explateid'=>$explateId));
		$imagingTasks=$stmt->fetchAll(PDO::FETCH_ASSOC);
		
	} else if (null!=$fromDate){

		//get the ImagingTasks
		//Time zone: See header comment for info on local versus GMT times in RockMaker. 
		//TODO $fromDate is not corrected for GMT
		//Just interested in preventing very recent inspections from importing, so comparing (local) DateImaged to (local) GETDATE() is good enough.
		$stmt=$rm->prepare('SELECT * FROM ImagingTask WHERE DateImaged>=:startdate AND DateImaged<=DATEADD(minute, -5, GETDATE()) ORDER BY DateImaged ASC');
		$stmt->execute(array(':startdate'=>$fromDate));
		$imagingTasks=$stmt->fetchAll(PDO::FETCH_ASSOC);
		
	} else if (null!=$imager){
		
		throw new Exception('Per-imager import not implemented yet');		
		
	} else {
	
		$highestImportedId=0;
		$lastImportDate=null;
		$limsSession=imagingsession::getAll(array(
				'sortby'=>'imageddatetime',
				'sortdescending'=>'yes',
				'pagesize'=>1,
				'pagenumber'=>1
		));
		if(isset($limsSession) && isset($limsSession['rows']) && !empty($limsSession['rows'])){
			$limsSession=$limsSession['rows'][0];
			$lastImportDate=$limsSession['imageddatetime'];
		}
		if(null==$lastImportDate){
			writeLog(LOGLEVEL_WARN, 'Cannot find latest inspection in LIMS.');
			writeLog(LOGLEVEL_WARN, 'Re-run with "from date" option, e.g, "-d20151231".');
			$imagingTasks=array();	
		} else {
			//Time zone: For preventing very recent inspections from importing, comparing (local) DateImaged to (local) GETDATE() is good enough.
			//However, we need to convert the GMT last imaged date in the LIMS to RockMaker database local time when deciding what to import.
			$stmt=$rm->prepare('SELECT * FROM ImagingTask 
					WHERE DATEADD(second, DATEDIFF(second, GETDATE(), GETUTCDATE()), DateImaged)>=:dateimaged 
					AND DateImaged<=DATEADD(minute, -5, GETDATE()) ORDER BY DateImaged ASC');
			$stmt->execute(array(':dateimaged'=>$lastImportDate));
			$imagingTasks=$stmt->fetchAll(PDO::FETCH_ASSOC);
		}
	}

	/*
	 * MSSQL "limit" is exceedingly messy. Take the performance hit, pull them all in above and chop it in PHP.
	 * In regular operation it is likely that far fewer than 10 rows will be returned anyway. This will
	 * only hurt when importing a lot of historical inspections.
	 */
	$imagingTasks=array_splice($imagingTasks, $limitStart, $limitTotal);

	if(empty($imagingTasks)){
		writeLog(LOGLEVEL_WARN, 'No imaging tasks to import');
	} else {
		foreach($imagingTasks as $it){
			$state=$it['State'];
			$taskId=$it['ID'];
			if($state==IMAGING_STATE_SKIPPED){
				writeLog(LOGLEVEL_WARN, 'Imaging task '.$taskId.' was Skipped, nothing to import');
				continue;
			} else if($state==IMAGING_STATE_CANCELLED){
				writeLog(LOGLEVEL_WARN, 'Imaging task '.$taskId.' was cancelled while in progress. It may or may not have images.');
				//don't break, try import anyway.
			} else if($state!=IMAGING_STATE_COMPLETED){
				writeLog(LOGLEVEL_INFO, 'Imaging task '.$taskId.' is not marked Completed, checking for later tasks on same imager');
				$barcode=getPlateBarcodeForImagingTask($taskId);
				$imagerName=getImagerNameForImagingTask($taskId, $barcode);
				if(IMAGER_NAME_UNKNOWN==$imagerName){
					writeLog(LOGLEVEL_WARN, 'Could not determine imager for this task. Will continue if task is more than 6 hours old.');
					$dateImaged=date_create_from_format('Y-m-d H:i:s', substr($id['DateImaged'], 0, 19));
					$diff = $dateImaged->diff(new Datetime('now'));
					if($diff->d >= 1 || $diff->h >= 6){
						writeLog(LOGLEVEL_WARN, 'Task is more than 6 hours old, so continuing.');
					} else {
						writeLog(LOGLEVEL_WARN, 'Task is less than 6 hours old, aborting import and stopping.');
						break; //Drop out of imaging tasks loop altogether.
					}
				} else {
					writeLog(LOGLEVEL_DEBUG, 'Imager for this task was '.$imagerName);
					$stmt=$ri->prepare('SELECT TOP 10 ImagingLog.ID, InspectionID, Robot.Name, EndTime
						FROM ImagingLog, Robot
						WHERE ImagingLog.RobotID=Robot.ID
						AND ImagingLog.EndTime>:dateimaged
						AND ImagingLog.InspectionID<>:taskid
						AND Robot.Name=:imagername
						order by Imaginglog.EndTime ASC'
					);
					$stmt->execute(array(':dateimaged'=>$it['DateImaged'], ':imagername'=>$imagerName, ':taskid'=>$taskId));
					$laterInspections=$stmt->fetchAll(PDO::FETCH_ASSOC);
					if(!empty($laterInspections)){
						//Robot did another plate after this one, therefore this one is finished
						writeLog(LOGLEVEL_WARN, 'Found later tasks on the same imager, so this one must be finished. Continuing.');
					} else {
						writeLog(LOGLEVEL_WARN, 'Could not find task(s) for this imager after the incomplete one, so aborting import and stopping.');
						break; //Drop out of imaging tasks loop altogether.
					}
				}
			}
			database::begin();
			importImagingTask($taskId);
			database::commit();
			writeLog(LOGLEVEL_INFO,'-------------------------------------------------');
		}
	}
	
	unset($rm); unset($ri); unset($stmt);

} catch(Exception $e){
	writeLog(LOGLEVEL_ERROR, get_class($e).' caught: '.$e->getMessage());
	$trace=$e->getTrace();
	foreach($trace as $t){
		writeLog(LOGLEVEL_ERROR, ': '.$t['file']);
		if(isset($t['type']) && isset($t['class'])){
			writeLog(LOGLEVEL_ERROR, ':    Line '.$t['line'].', Function '.$t['class'].$t['type'].$t['function']);
		} else {
			writeLog(LOGLEVEL_ERROR, ':    Line '.$t['line'].', Function '.$t['function']);
		}
		//TODO $t['args']
	}
	database::abort();
	writeLog(LOGLEVEL_ERROR, 'Import aborted due to exception');
}

writeLog(LOGLEVEL_INFO,'Importer finished.');
writeLog(LOGLEVEL_INFO,'=================================================');

/**
 * Creates one or more plate inspections in the LIMS, from the Formulatrix/RockMaker ImagingTask with the specified ID.
 * If the ImagingTask has more than one CaptureProfile (for example, both visible and UV images), one LIMS plate inspection
 * will be created for each, with the inspections slightly separated in time. This ensures that there is only one image per 
 * drop per time point in the LIMS. 
 * @param unknown $taskId The RockMaker ImagingTask ID
 * @return void
 */
function importImagingTask($taskId){
	global $rm, $ri, $sharedProjectId, $lightPaths;
	
	writeLog(LOGLEVEL_INFO, 'Importing imaging task, ID '.$taskId);
	
	$limsTask=imagingsession::getByProperty('manufacturerdatabaseid', $taskId);
	if($limsTask){
		//This is a re-import, assume the metadata is OK and just look for new images
		writeLog(LOGLEVEL_INFO, 'Imaging task with Formulatrix ID '.$taskId.' already imported, will re-import in case new images were written');
	}
	
	/*
	 * Get imaging task information.
	 * See header comment for info on local vs GMT in RockMaker.
	 */ 
	
	//$stmt=$rm->prepare('SELECT CONVERT(VARCHAR(19), ImagingTask.DateImaged, 126) AS DateImaged, 
	$stmt=$rm->prepare('SELECT CONVERT(VARCHAR(19), ImagingTask.DateImaged, 126) AS DateImagedLocal, 
			CONVERT(VARCHAR(19), DATEADD(second, DATEDIFF(second, GETDATE(), GETUTCDATE()), DateImaged), 126) AS DateImagedUtc,
			ImageBatch.ID AS ImageBatchID, Plate.ID AS PlateID, Plate.Barcode AS Barcode
			FROM ImagingTask, ImageBatch, ExperimentPlate, Plate 
			WHERE ExperimentPlate.ID=ImagingTask.ExperimentPlateID 
			AND ImageBatch.ImagingTaskID=ImagingTask.ID
			AND ExperimentPlate.PlateID=Plate.ID
			AND ImagingTask.ID=:taskid
	'); //CONVERT to varchar(19) to lop off the milliseconds. Date looks like 2011-12-13T12:42:55 after conversion.
	$stmt->execute(array(':taskid'=>$taskId));
	$imagingTask=$stmt->fetch(PDO::FETCH_ASSOC);
	if(!$imagingTask){
		writeLog(LOGLEVEL_WARN, 'No matching ImageBatch for ImagingTask '.$taskId.', so cannot locate images. Not importing this ImagingTask.');
		return false;
	}
	
	/*
	 * Are there any images for this ImagingTask? It appears that there are some, even marked as Completed, that do not have any. I don't know all the reasons
	 * why this can happen, but certainly in older versions of RockImager, if the USB stick containing the MIL (Matrox Imaging Library) license works loose 
	 * you'll see a plate get to the camera, a license exception thrown, and the ImagingTask marked as Completed.
	 * 
	 * We should check this first, before importing pre-requisites that may not be needed, and abort import of this ImagingTask if need be.
	 */  
	$stmt=$rm->prepare('SELECT ID FROM ImageType WHERE ShortName=\'ef\'');
	$stmt->execute(array());
	$result=$stmt->fetch(PDO::FETCH_ASSOC);
	$imageTypeId=$result['ID'];
	writeLog(LOGLEVEL_DEBUG, 'Image type ID: '.$imageTypeId);
	$stmt=$rm->prepare('SELECT ID FROM RegionType WHERE Name=\'Drop\'');
	$stmt->execute(array());
	$result=$stmt->fetch(PDO::FETCH_ASSOC);
	$regionTypeId=$result['ID'];
	writeLog(LOGLEVEL_DEBUG, 'Region type ID: '.$regionTypeId);
	$stmt=$rm->prepare('SELECT ImageStore.BasePath, ImageBatch.ID AS ImageBatchID, CaptureResult.ID AS CaptureResultID,
			CaptureProfileVersion.CaptureProfileID AS CaptureProfileID, CaptureProfileVersion.ID AS CaptureProfileVersionID,
			Region.ID AS RegionID, Plate.Barcode, Well.PlateID, Well.WellNumber, Well.RowLetter, Well.ColumnNumber, WellDrop.DropNumber,
			Image.PixelSize
		FROM Image, ImageStore, CaptureResult, CaptureProfileVersion, ImageBatch, ImagingTask, Region, WellDrop, Well, Plate
		WHERE Image.ImageTypeID=:imagetypeid
			AND Image.ImageStoreID=ImageStore.ID
			AND Image.CaptureResultID=CaptureResult.ID
			AND CaptureResult.ImageBatchID=ImageBatch.ID
			AND CaptureResult.RegionID=Region.ID
			AND Region.RegionTypeID=:regiontypeid
			AND CaptureResult.CaptureProfileVersionID=CaptureProfileVersion.ID
			AND Region.WellDropID=WellDrop.ID
			AND WellDrop.WellID=Well.ID
			AND Well.PlateID=Plate.ID
			AND ImageBatch.ImagingTaskID=ImagingTask.ID
			AND ImagingTask.ID=:taskid
		ORDER BY CaptureProfileID, Well.ID ASC
	');
	$stmt->execute(array(':taskid'=>$taskId, ':imagetypeid'=>$imageTypeId, ':regiontypeid'=>$regionTypeId));
	$images=$stmt->fetchAll(PDO::FETCH_ASSOC);
	if(empty($images)){
		writeLog(LOGLEVEL_WARN, 'No images found for ImagingTask ID  '.$taskId.', not importing it');
		return;
	}
	writeLog(LOGLEVEL_DEBUG, count($images).' images found for ImagingTask ID  '.$taskId);

	
	$fxPlateId=$imagingTask['PlateID'];
	$imageBatchId=$imagingTask['ImageBatchID'];
	$barcode=$imagingTask['Barcode'];
	$imagedTime=str_replace('T', ' ', $imagingTask['DateImagedUtc']);
	
	//Check whether the LIMS has the plate. If not, import it (and its plate type if needed).
	$limsPlate=plate::getByName($barcode);
	writeLog(LOGLEVEL_INFO, 'Plate barcode is '.$barcode.', time imaged '.$imagedTime);
	if($limsPlate){
		writeLog(LOGLEVEL_DEBUG, 'Plate '.$barcode.' exists in LIMS');
	} else {
		writeLog(LOGLEVEL_INFO, 'Plate '.$barcode.' does not exist in LIMS, creating it');
		importPlate($barcode);
	}

	//If plate STILL doesn't exist in the LIMS, bail.
	$limsPlate=plate::getByName($barcode);
	if(!$limsPlate){
		throw new Exception('Plate '.$barcode.' does not exist in LIMS after calling importPlate.');
	}

	//All records under this plate need to share its project ID. Imagers, plate types, etc., go under the Shared project.
	$projectId=$limsPlate['projectid'];

	$imagerName=getImagerNameForImagingTask($taskId, $barcode);

	$limsImager=imager::getbyName($imagerName);
	if(!$limsImager){
		writeLog(LOGLEVEL_INFO, 'Imager '.$imagerName.' does not exist in LIMS, creating it.');
		importImager($imagerName);
		$limsImager=imager::getbyName($imagerName);
	}
	if(!$limsImager){
		throw new Exception('Imager '.$imagerName.' does not exist in LIMS after calling importImager.');
	}
	
	
	//We want to split the Formulatrix ImagingTask into one LIMS imagingsession per capture profile, but do not want them to have the same 
	//timestamp in the LIMS. We subtract $CaptureProfileId seconds from the Formulatrix time when setting the LIMS imagingsession time - 
	//but NOT for the lowest profile ID. If we did the same with that ID, then an "import all ImagingTasks after the last imagingsession time"
	//job would re-import the most recent ImagingTask.
	$minimumProfileId=999999;
	foreach($images as $img){
		if($img['CaptureProfileID']<$minimumProfileId){
			$minimumProfileId=$img['CaptureProfileID'];
		}
	}
	
	$limsImagingSessions=array();
	$verifiedImageStores=array();
	$existingProfiles=array();
	$existingProfileVersions=array();

	/*
	 * Verify that the LIMS image store exists
	 */
	$limsStorage=config::get('imagestore');
	$limsStorage=rtrim($limsStorage,'/\\').'/';
	if(!in_array($limsStorage, $verifiedImageStores)){
		if(!file_exists($limsStorage)){
			writeLog(LOGLEVEL_ERROR, 'Cannot find LIMS image store: '.$limsStorage);
			writeLog(LOGLEVEL_ERROR, 'Check that the drive is mounted and has the correct permissions');
			throw new Exception('Aborting because LIMS image store cannot be read');
		}
		$verifiedImageStores[]=$limsStorage;
	}
	
	/*
	 * Process each image in turn - set up pre-requisites, copy files across, create database record in LIMS
	 */
	foreach($images as $img){
		writeLog(LOGLEVEL_DEBUG, 'Found image details for '.$img['RowLetter'].$img['ColumnNumber'].'.'.$img['DropNumber'].' CaptureProfile '.$img['CaptureProfileID']);
		$basePath=$img['BasePath'];
		$profileId=$img['CaptureProfileID'];
		$profileVersionId=$img['CaptureProfileVersionID'];
		$fxPlateId=$img['PlateID'];
		$barcode=$img['Barcode'];
		$imageBatchId=$img['ImageBatchID'];
		$path='/WellImages/'.($fxPlateId%1000).'/plateID_'.$fxPlateId.'/batchID_'.$imageBatchId;

		/*
		 * Verify that the Formulatrix image store for this image exists
		 */
		if(!in_array($basePath, $verifiedImageStores)){
			writeLog(LOGLEVEL_DEBUG, 'Checking Formulatrix image store exists...');
			if(!file_exists($basePath)){
				writeLog(LOGLEVEL_ERROR, 'Cannot find image store: '.$img['BasePath']);
				writeLog(LOGLEVEL_ERROR, 'Check that the drive is mounted and has the correct permissions');
				throw new Exception('Aborting because Formulatrix image store cannot be read');
			}
			$verifiedImageStores[]=$basePath;
			writeLog(LOGLEVEL_DEBUG, '...Formulatrix image store exists.'); 
		} else {
			writeLog(LOGLEVEL_DEBUG, 'Formulatrix image store already known to exist');
		}

		
		/*
		 * Create the CaptureProfile / imagingparameters in the LIMS if it doesn't exist
		 */
		if(!isset($existingProfiles['fx'.$profileId])){
			writeLog(LOGLEVEL_DEBUG, 'Checking whether CaptureProfile exists in LIMS...');
			$limsProfile=imagingparameters::getByProperty('manufacturerdatabaseid', $profileId);
			if(!$limsProfile){ 
				importCaptureProfile($profileId); 
				$limsProfile=imagingparameters::getByProperty('manufacturerdatabaseid', $profileId);
			}
			$limsProfile=$limsProfile['rows'][0];
			$existingProfiles['fx'.$profileId]=$limsProfile;
			writeLog(LOGLEVEL_DEBUG, '...CaptureProfile exists.'); 
		} else {
			writeLog(LOGLEVEL_DEBUG, 'CaptureProfile already known to exist in LIMS...');
		}
		
		writeLog(LOGLEVEL_DEBUG, 'Checking whether CaptureProfileVersion exists in LIMS...');
		/*
		 * Create the CaptureProfileVersion / imagingparametersversion in the LIMS if it doesn't exist
		 */
		if(isset($existingProfileVersions['fx'.$profileVersionId])){
			$limsProfileVersion=$existingProfileVersions['fx'.$profileVersionId];
		} else {
			$limsProfileVersion=imagingparametersversion::getByProperty('manufacturerdatabaseid', $profileVersionId);
			if(!$limsProfileVersion){ 
				importCaptureProfileVersion($profileVersionId); 
				$limsProfileVersion=imagingparametersversion::getByProperty('manufacturerdatabaseid', $profileVersionId);
			}
			$limsProfileVersion=$limsProfileVersion['rows'][0];
			$imagerSettings=imagingparametersversion::getimagersettings($limsProfileVersion['id']);
			
			/*
			 * Determine the light type for this imaging task
			 */ 
			writeLog(LOGLEVEL_DEBUG, 'Determining light type for this imaging session');
			$lightPathId=0; //Visible
			$lightPathSettings=imagingsetting::getByProperty('settingname', 'LightPath');
			if(empty($lightPathSettings) || empty($lightPathSettings['rows'])){
				writeLog(LOGLEVEL_DEBUG, 'No light path settings in LIMS database');
			} else {
				writeLog(LOGLEVEL_DEBUG, 'Found light path settings in LIMS database, iterating');
				foreach($lightPathSettings['rows'] as $s){
					$settingValue=$s['settingvalue'];					
					writeLog(LOGLEVEL_DEBUG, 'Setting has light path '.$settingValue);
					if($limsProfileVersion['id']==$s['imagingparametersversionid']){
						if($limsImager['id']==$s['imagerid']){
							writeLog(LOGLEVEL_DEBUG, 'Matched Robot, CaptureProfileVersion: LightPath '.$settingValue.' ['.$lightPaths[$settingValue].']');
							writeLog(LOGLEVEL_DEBUG, 'Breaking out of iteration');
							$lightPathId=$settingValue;
							break;
						} else if(IMAGER_NAME_UNKNOWN==$limsImager['name'] && $settingValue>=0){
							writeLog(LOGLEVEL_DEBUG, 'Matched (unknown imager), CaptureProfileVersion: LightPath '.$settingValue.' ['.$lightPaths[$settingValue].']');
							writeLog(LOGLEVEL_DEBUG, 'Breaking out of iteration');
							$lightPathId=$settingValue;
							break;
						}
					} else {
						writeLog(LOGLEVEL_DEBUG, 'Setting did not match CaptureProfileVersion, irrelevant');
					}
				}
			}
			if($lightPathId>=count($lightPaths)){
				throw new Exception('Unknown light path ID value'.$lightPathId.' is out of range.');
			} else if($lightPathId<0){
				writeLog(LOGLEVEL_WARN, 'Imager is not capable of light path specified in CaptureProfile. Assuming visible.');
				$lightPathId=0; //Visible;
			}
			$lightPath=$lightPaths[$lightPathId];			
			writeLog(LOGLEVEL_DEBUG, 'Light type was determined to be '.$lightPathId.' ['.$lightPaths[$lightPathId].']');

			$limsProfileVersion['lightpath']=$lightPath;
			$existingProfileVersions['fx'.$profileVersionId]=$limsProfileVersion;
		}
		writeLog(LOGLEVEL_DEBUG, '...CaptureProfileVersion exists.');
		
		/*
		 * Create the LIMS imagingsession if it doesn't exist already
		 */
		$limsImagingSession=null;
		if(isset($limsImagingSessions['profile'.$img['CaptureProfileID']])){
			$limsImagingSession=$limsImagingSessions['profile'.$img['CaptureProfileID']];
		} else {
			writeLog(LOGLEVEL_DEBUG, 'LIMS imagingsession for this image has not been found yet, looking for it');
			$captureProfileId=$img['CaptureProfileID'];
			$date=DateTime::createFromFormat('Y-m-d H:i:s',$imagedTime);
			if($minimumProfileId!=$captureProfileId){
				//subtract (captureProfileId) seconds from time, but not for lowest CaptureProfileId.
				$date=$date->sub(date_interval_create_from_date_string($captureProfileId.' sec'));
			}
			$imagingSessionName=$barcode.'_'.date_format($date, 'Y-m-d_h:i:s').'_profile'.$captureProfileId;
			$limsImagingSession=imagingsession::getByName($imagingSessionName);
			if(!empty($limsImagingSession)){
				writeLog(LOGLEVEL_DEBUG, 'LIMS imagingsession for this image was found');
			} else {
				writeLog(LOGLEVEL_DEBUG, 'LIMS imagingsession for this image was not found, creating it');
				$params=array(
						'name'=>$imagingSessionName,
						'manufacturerdatabaseid'=>$taskId,
						'imagerid'=>$limsImager['id'],
						'plateid'=>$limsPlate['id'],
						'imageddatetime'=>date_format($date, 'Y-m-d H:i:s'),
						'lighttype'=>$limsProfileVersion['lightpath'],
						'imagingparametersversionid'=>$limsProfileVersion['id']
				);
				imagingsession::create($params);
				$limsImagingSession=imagingsession::getByName($imagingSessionName);
			}
			if(!$limsImagingSession){
				throw new Exception('Could not find imagingsession in LIMS after create');
			}
			$limsImagingSessions['profile'.$captureProfileId]=$limsImagingSession;
			writeLog(LOGLEVEL_DEBUG, 'LIMS imagingsession created');
		}
		if(!$limsImagingSession){
			throw new Exception('Could not find LIMS imaging session for ImagingTask '.$taskId.' CaptureProfile '.$img['CaptureProfileID']);
		}
		
		/*
		 * Check that the extended-focus image and its thumbnail exist in the Formulatrix image store
		 */
		writeLog(LOGLEVEL_DEBUG, 'Checking for existence of EF and thumbnail in Formulatrix store');
		$fxBaseName=$basePath.$path.'/wellNum_'.$img['WellNumber'].'/profileID_'.$img['CaptureProfileID'].'/d'.$img['DropNumber'].'_r'.$img['RegionID'];
		$fxExtendedFocusPath=$fxBaseName.'_ef.jpg';
		$fxThumbnailPath=$fxBaseName.'_th.jpg';
		if(!file_exists($fxExtendedFocusPath)){
			writeLog(LOGLEVEL_WARN,'Extended-focus image not found at '.$fxExtendedFocusPath.', not importing image');
			continue;			
		}
		writeLog(LOGLEVEL_DEBUG, '...EF image exists.');
		
		if(!file_exists($fxThumbnailPath)){
			writeLog(LOGLEVEL_WARN,'Thumbnail image not found at '.$fxThumbnailPath.', not importing image');
			continue;
		}
		writeLog(LOGLEVEL_DEBUG, '...Thumbnail exists.');
		
		/*
		 * Create the directory of images from this imagingsession if it does not exist
		 */
		writeLog(LOGLEVEL_DEBUG, 'Checking for imagingsession directory in LIMS image store...');
		$destinationdir=substr($barcode,0,2).'/'.substr($barcode,0,3).'/'.$barcode.'/imagingsession'.$limsImagingSession['id'].'/';
		@mkdir($limsStorage.$destinationdir.'/thumbs',0755,true);
		if(!file_exists($limsStorage.$destinationdir.'/thumbs')){
			writeLog(LOGLEVEL_ERROR, 'Path does not exist: '.$destinationdir.'/thumbs');
			throw new Exception('Could not create destination directory in LIMS image store');
		}
		writeLog(LOGLEVEL_DEBUG, '...exists.');
		
		/*
		 * Copy the extended-focus image and its thumbnail to the LIMS image store
		 */
		writeLog(LOGLEVEL_DEBUG, 'Copying images to LIMS store');
		$efDestination=$destinationdir.$img['RowLetter'].str_pad($img['ColumnNumber'],2,'0',STR_PAD_LEFT).'_'.$img['DropNumber'].'.jpg';
		$thDestination=$destinationdir.'thumbs/'.$img['RowLetter'].str_pad($img['ColumnNumber'],2,'0',STR_PAD_LEFT).'_'.$img['DropNumber'].'.jpg';
		copy($fxExtendedFocusPath, $limsStorage.$efDestination);
		copy($fxThumbnailPath, $limsStorage.$thDestination);
			if(!file_exists($limsStorage.$efDestination)){
			writeLog(LOGLEVEL_ERROR, 'EF image not copied: '.$limsStorage.$efDestination);
			throw new Exception('Could not create EF image in LIMS image store');
		}
		if(!file_exists($limsStorage.$thDestination)){
			writeLog(LOGLEVEL_ERROR, 'Thumbnail image not copied: '.$limsStorage.$thDestination);
			throw new Exception('Could not create thumbnail in LIMS image store');
		}
		$imageSize=getimagesize($limsStorage.$efDestination);
		if(!$imageSize){
			writeLog(LOGLEVEL_WARN, ''.$limsStorage.$efDestination.' may be corrupt, could not get pixel dimensions');
			writeLog(LOGLEVEL_WARN, 'Not linking image '.$limsStorage.$efDestination.' in LIMS database');
		} else {
			$imageWidth=$imageSize[0];
			$imageHeight=$imageSize[1];
			writeLog(LOGLEVEL_DEBUG, 'Creating database record of image...');
			$imageDbName=$barcode.'_'.$img['RowLetter'].str_pad($img['ColumnNumber'],2,'0',STR_PAD_LEFT).'.'.$img['DropNumber'].'_is'.$limsImagingSession['id'];
			$preExisting=dropimage::getByName($imageDbName);
			if(!$preExisting){
				$welldropName=$barcode.'_'.$img['RowLetter'].str_pad($img['ColumnNumber'],2,'0',STR_PAD_LEFT).'.'.$img['DropNumber'];
				$welldrop=welldrop::getByName($welldropName);
				if(!$welldrop){
					throw new Exception('welldrop not found in LIMS for '.$welldropName);
				}
				dropimage::create(array(
						'name'=>$imageDbName,
						'projectid'=>$projectId,
						'imagingsessionid'=>$limsImagingSession['id'],
						'welldropid'=>$welldrop['id'],
						'pixelheight'=>$imageHeight,
						'pixelwidth'=>$imageWidth,
						'micronsperpixelx'=>$img['PixelSize'],
						'micronsperpixely'=>$img['PixelSize'],
						'imagestorepath'=>$limsStorage,
						'imagepath'=>$efDestination,
						'thumbnailpath'=>$thDestination
				));
			}
			writeLog(LOGLEVEL_DEBUG, '...done');
		}
		writeLog(LOGLEVEL_DEBUG, 'Copied image and thumbnail to LIMS store');


	} //End of images loop
	
	writeLog(LOGLEVEL_INFO, 'Imaging task imported');
}




/**
 * Creates a record in the LIMS of the imager with the specified serial/name.
 * Capacity calculations may be somewhat complicated by the existence in the Rigaku DB of a second hotel that is not installed.
 * @param string $imagerName The imager name.
 * @throws Exception
 */
function importImager($imagerName){
	writeLog(LOGLEVEL_DEBUG, 'In importImager, imagerName='.$imagerName);
	global $ri, $sharedProjectId;
	$imagerType=explode('-', $imagerName)[0];
	$capacity=10000;
	//TODO Calculate capacity from table structure
	//Looks like Hotels are swappable, Helsinki's has Linbro and SBS 
	
	imager::create(array(
			'name'=>$imagerName,
			'friendlyname'=>$imagerName,
			'manufacturer'=>'Rigaku',
			'temperature'=>20,
			'platecapacity'=>$capacity,
			'alertlevel'=>floor($capacity*IMAGERLOAD_ALERT_THRESHOLD_PERCENT/100),
			'warninglevel'=>floor($capacity*IMAGERLOAD_WARNING_THRESHOLD_PERCENT/100)
	));
	writeLog(LOGLEVEL_DEBUG, 'Returning from importImager');
	
}




/**
 * Creates an imagingparametersversion record in the LIMS of a Formulatrix CaptureProfileVersion, and
 * updates the parent imagingparameters to reflect the current version in the Formulatrix system.
 * WARNING This could fail if two versions of a CaptureProfile are created and used in quick succession. Importing the first causes 
 * the second to be set as current on the LIMS' copy of the CaptureProfile, likely causing a DB constraint violation. Realistically, if
 * the importer runs every 5-10 minutes and it takes that long to image a plate, this situation should not arise. Historical data may cause issues though.
 * @param unknown $captureProfileVersionId
 */
function importCaptureProfileVersion($captureProfileVersionId){
	writeLog(LOGLEVEL_DEBUG, 'In importCaptureProfileVersion, id='.$captureProfileVersionId);
	global $rm, $sharedProjectId;
	$stmt=$rm->prepare('SELECT CaptureProfileVersion.ID, CaptureProfileVersion.CaptureProfileID, CaptureProfile.CurrentCaptureProfileVersionID, CaptureProfile.Name
			FROM CaptureProfileVersion, CaptureProfile
			WHERE CaptureProfile.ID=CaptureProfileVersion.CaptureProfileID
			AND CaptureProfileVersion.ID=:id
	');	
	$stmt->execute(array(':id'=>$captureProfileVersionId));
	$fxVersion=$stmt->fetch(PDO::FETCH_ASSOC);
	if(!$fxVersion){
		throw new Exception('No CaptureProfileVersion in Formulatrix with ID '.$captureProfileVersionId);
	}
	
	$limsVersion=imagingparametersversion::getByProperty('manufacturerdatabaseid', $fxVersion['ID']);
	if($limsVersion){
		writeLog(LOGLEVEL_WARN, 'Called importCaptureProfileVersion but it already exists');
		return;
	}
	
	$limsProfile=imagingparameters::getByProperty('manufacturerdatabaseid', $fxVersion['CaptureProfileID']);
	if(!$limsProfile){
		importCaptureProfile($fxVersion['CaptureProfileID']);
	}
	$limsProfile=imagingparameters::getByProperty('manufacturerdatabaseid', $fxVersion['CaptureProfileID']);
	if(!$limsProfile){ throw new Exception('imagingparameters with manufacturerdatabaseid '.$fxVersion['CaptureProfileID'].' does not exist in LIMS'); }
	$limsProfile=$limsProfile['rows'][0];
	
	imagingparametersversion::create(array(
			'name'=>$fxVersion['Name'].'_v'.$fxVersion['CurrentCaptureProfileVersionID'],
			'imagingparametersid'=>$limsProfile['id'],
			'manufacturerdatabaseid'=>$captureProfileVersionId,
			'projectid'=>$sharedProjectId,
	));

	$limsVersion=imagingparametersversion::getByProperty('manufacturerdatabaseid', $fxVersion['ID']);
	if(!$limsVersion){
		throw new Exception('Could not create imagingparametersversion in LIMS');
	}
	$limsVersion=$limsVersion['rows'][0];
	
	//Update the parent imagingparameters with the current version (whether it's this version or not) - this is the part that may fail, see warning in function docs
	imagingparameters::update($limsVersion['id'], array(
			'currentversionid'=>$limsVersion['id']
	));
	
	//import the imager settings for this version of the profile
	$stmt=$rm->prepare('SELECT CaptureProperty.CaptureProfileVersionID, CaptureProperty.RobotID, CaptureProperty.Name AS Name, CaptureProperty.Value AS Value, Imager.Name AS RobotName
			FROM CaptureProperty, Imager
			WHERE CaptureProperty.RobotId=Imager.ID 
			AND CaptureProperty.CaptureProfileVersionID=:id
	');	
	$stmt->execute(array(':id'=>$captureProfileVersionId));
	$settings=$stmt->fetchAll(PDO::FETCH_ASSOC);
	foreach($settings as $s){	
		$imager=imager::getByName($s['RobotName']);
		if(!$imager){
			importImager($s['RobotName']);
			$imager=imager::getByName($s['RobotName']);
		}
		if(!$imager){
			throw new Exception('Could not find imager '.$s['RobotName'].' in LIMS');
		}
		imagingsetting::create(array(
				'settingname'=>$s['Name'],
				'settingvalue'=>$s['Value'],
				'imagerid'=>$imager['id'],
				'projectid'=>$sharedProjectId,
				'imagingparametersversionid'=>$limsVersion['id'],
				'name'=>'v'.$limsVersion['id'].'_robot'.$imager['id'].'_'.$s['Name']
		));
	}
	
	writeLog(LOGLEVEL_DEBUG, 'Returning from importCaptureProfileVersion');
}


/**
 * Creates an imagingparameters record in the LIMS of a Formulatrix CaptureProfile
 * @param unknown $captureProfileId
 */
function importCaptureProfile($captureProfileId){
	writeLog(LOGLEVEL_DEBUG, 'In importCaptureProfile, id='.$captureProfileId);
	global $rm, $sharedProjectId;
	$stmt=$rm->prepare('SELECT Name, CurrentCaptureProfileVersionID
			FROM CaptureProfile
			WHERE ID=:id
		');
	$stmt->execute(array(':id'=>$captureProfileId));
	$fxProfile=$stmt->fetch(PDO::FETCH_ASSOC);
	if(empty($fxProfile)){
		throw new Exception('RockMaker CaptureProfile with ID '.$captureProfileId.' does not exist');
	}
	
	$limsProfile=imagingparameters::getByName($fxProfile['Name']);
	if($limsProfile){
		writeLog(LOGLEVEL_WARN, 'Called importCaptureProfile but it already exists');
		return;
	}
	imagingparameters::create(array(
			'name'=>$fxProfile['Name'],
			'projectid'=>$sharedProjectId,
			'manufacturer'=>'Formulatrix',
			'manufacturerdatabaseid'=>$captureProfileId
	));
	$limsProfile=imagingparameters::getByName($fxProfile['Name']);
	if(!$limsProfile){
		throw new Exception('Could not create imagingparameters in LIMS');
	}	
	writeLog(LOGLEVEL_DEBUG, 'Returning from importCaptureProfile');
}

/**
 * Creates a record in the LIMS of the plate with the specified barcode.
 * If the plate type does not exist, importPlateType is called to create it.
 * @param unknown $barcode The plate barcode.
 * @return void
 */
function importPlate($barcode){
	writeLog(LOGLEVEL_DEBUG, 'In importPlate, barcode='.$barcode);
	global $rm, $scoringSystem, $unknownUsernames, $defaultPlateOwnerUsername;
	$stmt=$rm->prepare('SELECT Plate.ID, Containers.Name AS PlateTypeName, Users.Name AS UserName, Users.EmailAddress
			FROM Plate, Experiment, Containers, Users
			WHERE Plate.ExperimentID=Experiment.ID 
				AND Users.ID=Experiment.UserID 
				AND Containers.ID=ContainerID 
				AND Plate.Barcode=:barcode
		');
	$stmt->execute(array(':barcode'=>$barcode));
	$fxPlate=$stmt->fetch(PDO::FETCH_ASSOC);
	$plateTypeName=$fxPlate['PlateTypeName'];
	writeLog(LOGLEVEL_DEBUG, 'Plate '.$barcode.' has plate type: '.$plateTypeName);
	$limsPlateType=platetype::getByName($plateTypeName);
	if($limsPlateType){
		writeLog(LOGLEVEL_DEBUG, 'Plate type '.$plateTypeName.' exists in LIMS');
	} else {
		writeLog(LOGLEVEL_INFO, 'Plate type '.$plateTypeName.' does not exist in LIMS, creating it');
		importPlateType($plateTypeName);
		$limsPlateType=platetype::getByName($plateTypeName);
	}

	$project=project::getByName("Default Project");
	if(!$project){
		throw new Exception('Project "Default Project" does not exist.');
	}

	
	if(in_array($fxPlate['UserName'], $unknownUsernames)){
		writeLog(LOGLEVEL_WARN, 'Plate owner [' .$fxPlate['UserName']. '] should not own plates. Setting ['.$defaultPlateOwnerUsername.'] instead.');
		$plateOwner=user::getByName($defaultPlateOwnerUsername);
	} else {
		$plateOwner=user::getByProperty('email', $fxPlate['EmailAddress']);
		if(!$plateOwner){
			writeLog(LOGLEVEL_INFO, 'Plate owner [' .$fxPlate['EmailAddress']. '] does not exist in LIMS. Creating.');
			$plateOwner=importUser($fxPlate['EmailAddress']);
			if(!$plateOwner){
				throw new Exception('Could not create user in LIMS');
			}
		} else {
			$plateOwner=$plateOwner['rows'][0];
		}
	}
	$request=array(
			'name'=>$barcode, 
			'projectid'=>$project['id'],
			'platetypeid'=>$limsPlateType['id'],
			'ownerid'=>$plateOwner['id'],
			'crystalscoringsystemid'=>$scoringSystem['id']
	);
	plate::create($request);
	$limsPlate=plate::getByName($barcode);
	
	writeLog(LOGLEVEL_DEBUG, 'Returning from importPlate after create');
}

/**
 * Creates the Formulatrix scoring system in the LIMS, if it does not exist
 */
function importScoringSystem(){
	global $rm, $sharedProjectId;
	writeLog(LOGLEVEL_DEBUG, 'In importScoringSystem');
	
	$existing=crystalscoringsystem::getByName('Formulatrix');
	if($existing){ return $existing; }
	
	$system=crystalscoringsystem::create(array(
		'name'=>'Formulatrix'
	));
	if(!$system){ throw new Exception('Could not create scoring system'); }
	$system=$system['created'];
	
	global $rm, $ri, $sharedProjectId, $lightPaths;

	$stmt=$rm->prepare('SELECT WellDropScore.*
			FROM WellDropScore, WellDropScoreGroup
			WHERE WellDropScore.WellDropScoreGroupID=WellDropScoreGroup.ID
			AND WellDropScoreGroup.Name=:name
			ORDER BY OrderNumber ASC
	');
	$stmt->execute(array(':name'=>'Default'));
	$scores=$stmt->fetchAll(PDO::FETCH_ASSOC);
	$index=0;
	foreach($scores as $s){
		$color=substr("000000".dechex(1*$s['Color']),-6);
		$request=array(
				'crystalscoringsystemid'=>$system['id'],
				'scoreindex'=>$index,
				'color'=>$color,
				'label'=>$s['Name'],
				'name'=>'Formulatrix_'.$s['Name'],
		);
		crystalscore::create($request);
		$index++;
	}
	writeLog(LOGLEVEL_DEBUG, 'returning from importScoringSystem');
	return $system;
}



/**
 * Creates a record of the plate type in the LIMS, from the RockMaker Containers entry with the same name.
 * @param unknown $plateTypeName The name of the plate type
 */
function importPlateType($plateTypeName){
	writeLog(LOGLEVEL_DEBUG, 'In importPlateType, plateTypeName='.$plateTypeName);
	global $rm;

	$stmt=$rm->prepare('SELECT NumRows, NumColumns, MaxNumDrops
			FROM Containers WHERE Name=:name');
	$stmt->execute(array(':name'=>$plateTypeName));
	$plateType=$stmt->fetch(PDO::FETCH_ASSOC);

	//All plate types belong in the Shared project
	$sharedProject=project::getByName('Shared');
	
	//Make a default drop mapping with the drops along the top of the well (e.g., 123,RRR).
	//TODO Try to determine from RockImager calibration data
	$dropMapping=array('','');
	for($i=1;$i<=$plateType['MaxNumDrops'];$i++){
		$dropMapping[0].=$i;
		$dropMapping[1].='R';
	}
	$dropMapping=implode(',', $dropMapping);
	
	//Create the plate type in the LIMS
	$created=platetype::create(array(
		'name'=>$plateTypeName,
		'rows'=>$plateType['NumRows'],
		'cols'=>$plateType['NumColumns'],
		'subs'=>$plateType['MaxNumDrops'],
		'dropmapping'=>$dropMapping,
		'projectid'=>$sharedProject['id'],
	));
	writeLog(LOGLEVEL_DEBUG, 'Returning from importPlateType');
}



/**
 * Creates a record of a Rigaku user in the LIMS.
 * The user is created in an inactive state, with a default username, and will be unable to log in
 * until an administrator activates their account.
 * @param string $emailAddress The user's email address, registered in the Rigaku database
 */
function importUser($emailAddress){
	global $rm;
	$stmt=$rm->prepare('SELECT FIRST_NAME, LAST_NAME FROM ACCESS_USERS WHERE EMAIL=:email');
	$stmt->execute(array(':email'=>$emailAddress));
	$user=$stmt->fetch(PDO::FETCH_ASSOC);
	if(!$user){
		throw new Exception('Could not find Rigaku user with email address '.$emailAddress);
	}
	$userFullName=$user['FIRST_NAME'].' '.$user['LAST_NAME'];
	$temporaryUsername='AUTO'.rand(10000000,99999999);
	while(user::getByName($temporaryUsername)){
		writeLog(LOGLEVEL_DEBUG, 'Tried temporary username '.$temporaryUsername.' but it exists');
		$temporaryUsername='AUTO'.rand(10000000,99999999);
	}
	writeLog(LOGLEVEL_DEBUG, 'Using temporary username '.$temporaryUsername);
	$user=user::create(array(
			'name'=>$temporaryUsername,
			'fullname'=>$userFullName,
			'email'=>$emailAddress,
			'password'=>'USELESSPASSWORD',
			'isactive'=>0
	));
	return $user['created'];
}


function getPlateBarcodeForImagingTask($taskId){
	global $rig;
	$stmt=$rig->prepare('SELECT Plate.Barcode AS Barcode 
			FROM ImagingTask, ExperimentPlate, Plate
			WHERE ImagingTask.ExperimentPlateID=ExperimentPlate.ID
			AND ExperimentPlate.PlateID=Plate.ID
			AND ImagingTask.ID=:taskid');
	$stmt->execute(array(':taskid'=>$taskId));
	$task=$stmt->fetch(PDO::FETCH_ASSOC);
	if($task){
		return $task['Barcode'];
	}
	return null;
}

function getImagerNameForImagingTask($taskId, $barcode){
	global $ri,$sharedProject;
	
	$imagerName=null;
	//TODO Is there only one imager? Use it.
	
	//Check RockImager's imaging log for this specific imaging task
	if(!$imagerName){
		$stmt=$ri->prepare('SELECT Robot.Name AS RobotName
				FROM ImagingLog, Robot
				WHERE ImagingLog.RobotID=Robot.ID
					AND ImagingLog.PlateBarcode=:barcode
					AND ImagingLog.InspectionId=:taskid
		');
		$stmt->execute(array(':barcode'=>$barcode, ':taskid'=>$taskId));
		$logEntry=$stmt->fetch(PDO::FETCH_ASSOC);
		if($logEntry){
			$imagerName=$logEntry['RobotName'];
		}
	}
	//Then see if the plate is currently in an imager
	//TODO prioritise bigger imagers, since RI2 inventory can be out of date
	if(!$imagerName){
		$stmt=$ri->prepare('SELECT Robot.Name AS RobotName
				FROM Plate, Address, PlateAddress, Robot
				WHERE Address.RobotID=Robot.ID
					AND PlateAddress.AddressID=Address.ID
					AND PlateAddress.PlateID=Plate.ID
					AND Plate.Barcode=:barcode
		');
		$stmt->execute(array(':barcode'=>$barcode));
		$address=$stmt->fetch(PDO::FETCH_ASSOC);
		if($address){
			$imagerName=$address['RobotName'];
		}
	}
	//Has the plate been imaged before? Assume the same imager. Check the imaging log again,
	//without asking for a specific task ID, and take the most recent log entry.
	if(!$imagerName){
		$stmt=$ri->prepare('SELECT Robot.Name AS RobotName
			FROM ImagingLog, Robot
			WHERE ImagingLog.RobotID=Robot.ID
				AND ImagingLog.PlateBarcode=:barcode
			ORDER BY ImagingLog.InspectionID DESC
		');
		$stmt->execute(array(':barcode'=>$barcode));
		$logEntry=$stmt->fetch(PDO::FETCH_ASSOC);
		if($logEntry){
			$imagerName=$logEntry['RobotName'];
		}
	
	}
	//Use a default imager and let the user sort it out later. If the default doesn't exist, we create it.
	if(!$imagerName){
		writeLog(LOGLEVEL_WARN, 'Could not determine imager for ImagingTask '.$taskId.' on plate '.$barcode.', using default');
		$imagerName=IMAGER_NAME_UNKNOWN;
		$limsImager=imager::getByName($imagerName);
		if(!$limsImager){
			writeLog(LOGLEVEL_WARN, 'Default imager does not exist in LIMS, creating it');
			$sharedProject=project::getByName('Shared');
			imager::create(array(
					'name'=>$imagerName,
					'friendlyname'=>'Correct imager unknown',
					'manufacturer'=>'Unknown',
					'temperature'=>20,
					'platecapacity'=>100000,
					'alertlevel'=>100000,
					'warninglevel'=>100000,
					'projectid'=>$sharedProject['id']
			));
		}
	}
	writeLog(LOGLEVEL_DEBUG, 'Imager name is '.$imagerName);
	return $imagerName;
}


/**
 * Echoes a log message to the output, prefixed by a level indicator and date/time.
 * @param int $level The log level, see top of this file for definitions
 * @param string $message The log message
 */

function writeLog($level, $message){
	global $logLevel;
	if($level<$logLevel){ return; }
	$labels=array(
		LOGLEVEL_DEBUG=>'DEBUG',
		LOGLEVEL_INFO =>'INFO ',
		LOGLEVEL_WARN =>'WARN ',
		LOGLEVEL_ERROR=>'ERROR',
	);
	$out=$labels[$level].' '.date('d M y H:i:s').' '.$message."\n";
	echo $out;
}

function setUpClassAutoLoader(){
	spl_autoload_register(function($className){
		$paths=array(
				'../classes/',
				'../classes/core/',
				'../classes/core/exception/',
				'../classes/core/authentication/',
				'../classes/core/interface/',
				'../classes/model/',
		);
		foreach($paths as $path){
			if(file_exists($path.$className.'.class.php')){
				include_once($path.$className.'.class.php');
			}
		}
	});
}


function showHelp(){
	echo "\nImporter help\n\n";
	echo "By default this will import the ".IMAGINGSESSIONS_PER_RUN." imaging sessions after the most recent in the LIMS database, in date/time order. If none are present in the LIMS, the first ones in the Formulatrix database will be imported.\n";
	echo "\nRockImagerProcessor can lag some way behind the imager itself, meaning that JPEG images may not be available for some minutes after the imspection is shown as comple in RockImager. For this reason, no inspections younger than ".MINIMUM_INSPECTION_AGE_MINS." minutes old are imported.\n";
	echo "\nThe following arguments can be supplied to modify this behaviour:\n";
	echo "\n -h Show help";
	echo "\n -s5 - Skip this many inspections at the beginning (default 0)";
	echo "\n -m20 - Import at most this many inspections (default 10)";
	echo "\n -d2016-02-28 - Begin import from this date (default date of last imported inspection) - not compatible with other options";
	echo "\n -iRI54-0039 - Import only for this imager (default all imagers) - not compatible with other options";
	echo "\n -b909s - (Re)import all inspections and images for this barcode - not compatible with other options";
	echo '\n\nNote that the default behaviour is to import from the latest previously-imported imaging session. If the importer is running some way behind the imagers, using some of these options may mean that some imaging sessions are NEVER imported automatically.';
	exit;	
}

?>